@@ -1,20 +1,22 @@
 package Archive::Tar::Stream;
 
-use 5.006;
 use strict;
 use warnings;
 
 # this is pretty fixed by the format!
 use constant BLOCKSIZE => 512;
 
+use constant BLOCKCOUNT => 2048;
+use constant BUFSIZE => (512*2048);
+
 # dependencies
 use IO::File;
 use IO::Handle;
 use File::Temp;
+use List::Util qw(min);
 
 # XXX - make this an OO attribute
 our $VERBOSE = 0;
-our $NODIE = 0;
 
 =head1 NAME
 
@@ -22,11 +24,11 @@ Archive::Tar::Stream - pure perl IO-friendly tar file management
 
 =head1 VERSION
 
-Version 0.03
+Version 0.04
 
 =cut
 
-our $VERSION = '0.03';
+our $VERSION = '0.04';
 
 
 =head1 SYNOPSIS
@@ -84,7 +86,6 @@ Args:
    inpos     - initial offset in infh
    outpos    - initial offset in outfh
    safe_copy - boolean.
-   no_die    - boolean.
 
 Offsets are for informational purposes only, but can be
 useful if you are tracking offsets of items within your
@@ -104,45 +105,16 @@ file will not corrupt the output file.
 sub new {
   my $class = shift;
   my %args = @_;
-
-  unless ($args{infh} or $args{outfh}) {
-    if($NODIE) {
-      warn "Useless to stream without a filehandle";
-      return undef;
-    } else {
-      die "Useless to stream without a filehandle";
-    }
-  }
-
   my $Self = bless {
     # defaults
     safe_copy => 1,
-    no_die => 0,
     inpos => 0,
     outpos => 0,
     %args
   }, ref($class) || $class;
-
   return $Self;
 }
 
-=head2 NoDie
-
-   $ts->NoDie(1);
-
-Toggle the "no_die" field mentioned above (stops you suffering a die() on errors encountered - it's your responsability to not corrupt your own files if you do this).
-
-=cut
-
-sub NoDie {
-  my $Self = shift;
-  if (@_) {
-    $Self->{no_die} = shift;
-    $NODIE=$Self->{no_die};
-  }
-  return $Self->{no_die};
-}
-
 =head2 SafeCopy
 
    $ts->SafeCopy(0);
@@ -202,14 +174,9 @@ sub AddFile {
   my $size = shift;
   my $fh = shift;
 
-  my $header = $Self->BlankHeader(@_);
-  $header->{name} = $name;
-  $header->{size} = $size;
+  my $header = $Self->BlankHeader(@_, name => $name, size => $size);
 
-  my $fullheader = $Self->WriteHeader($header);
-  $Self->CopyFromFh($fh, $size);
-
-  return $fullheader;
+  return $size ? $Self->WriteFromFh($fh, $header) : $Self->WriteHeader($header);
 }
 
 =head2 AddLink
@@ -229,10 +196,7 @@ sub AddLink {
   my $name = shift;
   my $linkname = shift;
 
-  my $header = $Self->BlankHeader(@_);
-  $header->{name} = $name;
-  $header->{linkname} = $linkname;
-  $header->{typeflag} = 2;
+  my $header = $Self->BlankHeader(typeflag => 2, @_, name => $name, linkname => $linkname);
 
   return $Self->WriteHeader($header);
 }
@@ -291,7 +255,6 @@ sub StreamCopy {
 
   while (my $header = $Self->ReadHeader()) {
     my $pos = $header->{_pos};
-    my $oldsize = $header->{size};
     if ($Chooser) {
       my ($rc, $newheader) = $Chooser->($header, $Self->{outpos}, undef);
 
@@ -316,19 +279,15 @@ sub StreamCopy {
       if ($rc eq 'KEEP') {
         print "KEEP $header->{name} $pos/$Self->{outpos}\n" if $VERBOSE;
         if ($TempFile) {
-          $Self->WriteHeader($header);
-          $Self->CopyFromFh($TempFile, $header->{size});
+          $Self->WriteFromFh($TempFile, $header);
         }
-        # if the sizes don't match we just do a tempfile too
-        elsif ($Self->{safe_copy} or ($oldsize != $header->{size})) {
-          # guarantee safety by getting everything into a temporary file first
-          $TempFile = $Self->CopyToTempFile($oldsize);
-          $Self->WriteHeader($header);
-          $Self->CopyFromFh($TempFile, $header->{size});
+        # guarantee safety by getting everything into a temporary file first
+        elsif ($Self->{safe_copy} and $header->{size}) {
+          $TempFile = $Self->CopyToTempFile($header->{size});
+          $Self->WriteFromFh($TempFile, $header);
         }
         else {
-          $Self->WriteHeader($header);
-          $Self->CopyBytes($header->{size});
+          $Self->WriteCopy($header);
         }
       }
 
@@ -345,26 +304,17 @@ sub StreamCopy {
       }
 
       else {
-	if($NODIE) {
-          warn "Bogus response $rc from callback\n";
-	  return undef;
-        } else {
-          die "Bogus response $rc from callback\n";
-        }
+        die "Bogus response $rc from callback\n";
       }
     }
     else {
       print "PASSTHROUGH $header->{name} $Self->{outpos}\n" if $VERBOSE;
+      # XXX - faster but less safe
+      #$Self->WriteCopy($header);
 
-      if ($Self->{safe_copy}) {
-        my $TempFile = $Self->CopyToTempFile($header->{size});
-        $Self->WriteHeader($header);
-        $Self->CopyFromFh($TempFile, $header->{size});
-      }
-      else {
-        $Self->WriteHeader($header);
-        $Self->CopyBytes($header->{size});
-      }
+      # slow safe option :)
+      my $TempFile = $Self->CopyToTempFile($header->{size});
+      $Self->WriteFromFh($TempFile, $header);
     }
   }
 }
@@ -390,33 +340,23 @@ sub ReadBlocks {
   my $Self = shift;
   my $nblocks = shift || 1;
   unless ($Self->{infh}) {
-    if($NODIE) {
-      warn "Attempt to read without input filehandle";
-      return undef;
-    } else {
-      die "Attempt to read without input filehandle";
-    }
+    die "Attempt to read without input filehandle\n";
   }
   my $bytes = BLOCKSIZE * $nblocks;
-  my $buf;
-  my @result;
-  while ($bytes > 0) {
-    my $n = sysread($Self->{infh}, $buf, $bytes);
+  my $buf = '';
+  my $pos = 0;
+  while ($pos < $bytes) {
+    my $chunk = min($bytes - $pos, BUFSIZE);
+    my $n = sysread($Self->{infh}, $buf, $chunk, $pos);
     unless ($n) {
       delete $Self->{infh};
-      return if ($bytes == BLOCKSIZE * $nblocks); # nothing at EOF
-      if($NODIE) {
-	warn "Failed to read full block at $Self->{inpos}";
-	return undef;
-      } else {
-	die "Failed to read full block at $Self->{inpos}";
-      }
+      return unless $pos; # nothing at EOF is fine
+      die "Failed to read full block at $Self->{inpos}\n";
     }
-    $bytes -= $n;
+    $pos += $n;
     $Self->{inpos} += $n;
-    push @result, $buf;
   }
-  return join('', @result);
+  return $buf;
 }
 
 =head2 WriteBlocks
@@ -437,41 +377,29 @@ Returns the position of the header in the output stream.
 
 sub WriteBlocks {
   my $Self = shift;
-  my $string = shift;
+  my $buf = shift;
   my $nblocks = shift || 1;
 
   my $bytes = BLOCKSIZE * $nblocks;
 
   unless ($Self->{outfh}) {
-    if($NODIE) {
-      warn "Attempt to write without output filehandle";
-      return undef;
-    } else {
-      die "Attempt to write without output filehandle";
-    }
+    die "Attempt to write without output filehandle\n";
   }
   my $pos = $Self->{outpos};
 
   # make sure we've got $nblocks times BLOCKSIZE bytes to write
-  if (length($string) > $bytes) {
-    $string = substr($string, 0, $bytes);
-  }
-  elsif (length($string) < $bytes) {
-    $string .= "\0" x ($bytes - length($string));
+  if (length($buf) < $bytes) {
+    $buf .= "\0" x ($bytes - length($buf));
   }
 
-  while ($bytes > 0) {
-    my $n = syswrite($Self->{outfh}, $string, $bytes, (BLOCKSIZE * $nblocks) - $bytes);
+  my $bufpos = 0;
+  while ($bufpos < $bytes) {
+    my $n = syswrite($Self->{outfh}, $buf, $bytes - $bufpos, $bufpos);
     unless ($n) {
       delete $Self->{outfh};
-      if($NODIE) {
-	warn "Failed to write full block at $Self->{outpos}";
-	return undef;
-      } else {
-	die "Failed to write full block at $Self->{outpos}";
-      }
+      die "Failed to write full block at $Self->{outpos}\n";
     }
-    $bytes -= $n;
+    $bufpos += $n;
     $Self->{outpos} += $n;
   }
 
@@ -619,7 +547,6 @@ in %extra to make sure they're all known keys.
 
 sub BlankHeader {
   my $Self = shift;
-
   my %hash = (
     name => '',
     mode => 0777,
@@ -635,7 +562,6 @@ sub BlankHeader {
     devminor => 0,
     prefix => '',
   );
-
   my %overrides = @_;
   foreach my $key (keys %overrides) {
     if (exists $hash{$key}) {
@@ -729,8 +655,7 @@ sub DumpBytes {
   my $Self = shift;
   my $bytes = shift;
   while ($bytes > 0) {
-    my $n = int($bytes / BLOCKSIZE);
-    $n = 16  if $n > 16;
+    my $n = min(1 + int(($bytes-1) / BLOCKSIZE), BLOCKCOUNT);
     my $dump = $Self->ReadBlocks($n);
     $bytes -= length($dump);
   }
@@ -750,11 +675,9 @@ files together.
 
 sub FinishTar {
   my $Self = shift;
-  $Self->WriteBlocks("\0" x 512);
-  $Self->WriteBlocks("\0" x 512);
-  $Self->WriteBlocks("\0" x 512);
-  $Self->WriteBlocks("\0" x 512);
-  $Self->WriteBlocks("\0" x 512);
+  # add 5 blocks of all zero - this seems to be expected by gnutar and it will complain
+  # if they're not there
+  $Self->WriteBlocks("", 5);
 }
 
 =head2 CopyToTempFile
@@ -773,9 +696,9 @@ sub CopyToTempFile {
 
   my $TempFile = File::Temp->new();
   while ($bytes > 0) {
-    my $n = 1 + int(($bytes - 1) / BLOCKSIZE);
-    $n = 16  if $n > 16;
+    my $n = min(1 + int(($bytes - 1) / BLOCKSIZE), BLOCKCOUNT);
     my $dump = $Self->ReadBlocks($n);
+    die "Failed to read $n blocks for $bytes at $Self->{inpos}\n" unless defined $dump;
     $dump = substr($dump, 0, $bytes) if length($dump) > $bytes;
     $TempFile->print($dump);
     $bytes -= length($dump);
@@ -798,30 +721,96 @@ sub CopyFromFh {
   my $Self = shift;
   my $Fh = shift;
   my $bytes = shift;
+  my $buf = shift // '';
+  my $pos = shift // 0;
 
-  my $buf;
-  while ($bytes > 0) {
-    my $thistime = $bytes > BLOCKSIZE ? BLOCKSIZE : $bytes;
-    my $block = '';
-    while ($thistime) {
-      my $n = sysread($Fh, $buf, $thistime);
+  my $tocopy = $bytes + $pos;
+
+  while ($tocopy) {
+    my $chunk = min($tocopy - $pos, BUFSIZE);
+    if ($chunk) {
+      my $n = sysread($Fh, $buf, $chunk, $pos);
       unless ($n) {
-        if($NODIE) {
-	  warn "Failed to read entire file, doh ($bytes remaining)!\n";
-	  return undef;
-	} else {
-	  die "Failed to read entire file, doh ($bytes remaining)!\n";
-	}
+        die "Failed to read $chunk bytes from input fh at at $pos\n";
       }
-      $thistime -= $n;
-      $block .= $buf;
+      $pos += $n;
     }
-    if (length($block) < BLOCKSIZE) {
-      $block .= "\0" x (BLOCKSIZE - length($block));
+
+    # if we're done, write including padding
+    if ($pos == $tocopy) {
+      my $nblocks = 1 + int(($pos-1) / BLOCKSIZE);
+      $Self->WriteBlocks($buf, $nblocks);
+      return;
+    }
+
+    # if we have any full blocks, write them out
+    my $nblocks = int($pos / BLOCKSIZE);
+    if ($nblocks) {
+      $Self->WriteBlocks($buf, $nblocks);
+      # keep any partial blocks
+      my $written = $nblocks * BLOCKSIZE;
+      $buf = substr($buf, $written);
+      $pos -= $written;
+      $tocopy -= $written;
     }
-    $Self->WriteBlocks($block);
-    $bytes -= BLOCKSIZE;
   }
+
+  die "Finished copying without writing everything\n";
+}
+
+=head2 WriteFromFh
+
+   $ts->WriteFromFh($fh, $header);
+
+Adds the header and then calls CopyFromFh with the data from the
+filehandle
+
+=cut
+
+sub WriteFromFh {
+  my $Self = shift;
+  my $Fh = shift;
+  my $header = shift;
+
+  my $pos = $Self->{outpos};
+
+  my $block = $Self->CreateHeader($header);
+  $Self->CopyFromFh($Fh, $header->{size}, $block, BLOCKSIZE);
+
+  return( {%$header, _pos => $pos} );
+}
+
+=head2 WriteCopy
+
+   $ts->WriteCopy($header);
+
+Streams the record which matches the given header directly from the input
+stream to the output stream.
+
+=cut
+
+sub WriteCopy {
+  my $Self = shift;
+  my $header = shift;
+
+  my $pos = $Self->{outpos};
+
+  my $block = $Self->CreateHeader($header);
+  my $hn = 1;
+
+  my $bytes = $header->{size};
+  while ($hn || $bytes > 0) {
+    my $n = min(1 + int(($bytes-1) / BLOCKSIZE), BLOCKCOUNT);
+    my $dump = $n ? $Self->ReadBlocks($n) : '';
+    die "Failed to read $n blocks for $bytes at $Self->{inpos}\n" unless defined $dump;
+    $Self->WriteBlocks($block . $dump, $hn + $n);
+    $bytes -= length($dump);
+    # only write header on the first time;
+    $block = '';
+    $hn = 0;
+  }
+
+  return( {%$header, _pos => $pos} );
 }
 
 =head1 TARHEADER format
